{
 "cells": [
  {
   "attachments": {},
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Chapter 7: Binary Heap\n",
    "\n",
    "## Trees in Arrays\n",
    "The previous lectures store a binary tree with the help of pointer-liked structures, in which each item contains references to its children. \n",
    "If the tree is a `complete binary tree`, there is a useful `array-based` alternative.  \n",
    "\n",
    "**Definition**. A binary tree is complete if every level, except possibly the last, is completely filled, and all the leaves on the last level are placed as far to the left as possible.\n",
    "\n",
    "A complete binary tree is one that can be obtained by filling the nodes starting with the root, and then each next level in turn, always from the left, until one runs out of nodes. \n",
    "Complete binary trees always have minimal height for their size $n$, namely $logn$, and are always perfectly balanced.\n",
    "\n",
    "- where to store in an array P, \n",
    "    - root is at $0$\n",
    "    - left(i) = $2i+1$\n",
    "    - right(i) = $2i+2$\n",
    "    - parent(i) = $floor(\\frac{i-1}{2})$\n",
    "\n",
    "- Storing a binary tree as an array is not efficient \n",
    "    - if the tree is not complete, reserve space in the array for every possible node in the tree\n",
    "    - for binary search tree, insertion or deletion will involve shifting large portions of the array\n",
    "\n"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "\n",
    "## Heaps\n",
    "\n",
    "The (binary) heap data structure is an array object that we can view as a nearly complete binary tree.\n",
    "  - Each node of the tree corresponds to an element of the array. \n",
    "  - The tree is completely filled on all levels except possibly the lowest, which is filled from the left up to a point\n",
    "    - An array A that represents a heap is an object with two attributes: `A.length`, which (as usual) gives the number of elements in the array, and `A.size`, which represents how many elements in the heap are stored within array `A`.\n",
    "    - Given the index `i`, we can easily calculate its parent, left and right child based on the way they are stored.\n",
    "\n",
    "Max-heap:\n",
    "  - `max-heap property`: for every node `i` except the root, $A[parent(i)] >= A[i]$\n",
    "  - `max-heap` is an array satisfying max-heap property at all nodes\n",
    "  - **Min-heap** is symmetric\n",
    "\n"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Heap Operations\n",
    "\n",
    "```Python\n",
    "import math\n",
    "class P():\n",
    "    def __init__(self, n):\n",
    "        self.length = n\n",
    "        self.size = 0\n",
    "        self.A = [0]*n\n",
    "    \n",
    "    def left(self, i):\n",
    "        return 2*i + 1\n",
    "    \n",
    "    def right(self, i):\n",
    "        return 2*i + 2\n",
    "    \n",
    "    def parent(self, i):\n",
    "        return math.floor((i-1)/2)\n",
    "    \n",
    "    # The height of the tree is $floor(log(n))$. The height for each node $i$ is $floor(log(n)) - floor(log(i+1))$\n",
    "    def getHeight(self, i):\n",
    "        return math.floor(math.log(n)) - math.floor(math.log(i+1))\n",
    "    \n",
    "    # maintain the max-heap property of the `i`th node in heap `P` in $O(log(n))$ time from top to down\n",
    "    # check whether A[i] >= A[j] for j in {left(i), right(i)}\n",
    "    #    - if not, swap A[i] with A[j] for child j in {left(i), right(i)} with maximum value and recursively `heapifyDown(A, j)`.\n",
    "    def isValid(self, parent, left, right):\n",
    "        return self.A[parent] > self.A[left] and self.A[parent] > self.A[right]\n",
    "\n",
    "    def heapifyDown(self, i):\n",
    "        left = self.left(i)\n",
    "        right = self.right(i)\n",
    "        \n",
    "        if not self.isValid(i, left, right)\n",
    "            if self.isValid(left, i, right):\n",
    "                self.A[i], self.A[left] = self.A[left], self.A[i]\n",
    "                j = left\n",
    "            elif: self.isValid(right, left, i):\n",
    "                self.A[i], self.A[right] = self.A[right], self.A[i]\n",
    "                j = right\n",
    "        \n",
    "            self.heapifyDown(j)\n",
    "    # maintain the max-heap property of the `i`th node in heap `P` in $O(log(n))$ time from bottom to up\n",
    "    # check whether A[i] >= A[j] for j in {left(i), right(i)}\n",
    "    #    - if not, swap A[i] with A[j] for child j in {left(i), right(i)} with maximum value and recursively `heapifyDown(A, j)`.\n",
    "    def heapifyUp(self, i):\n",
    "        parent = self.parent(i)\n",
    "        \n",
    "        if self.A[i] > self.A[parent]:\n",
    "            self.A[i], self.A[parent] = self.A[parent], self.A[i]\n",
    "            j = parent\n",
    "            self.heapifyUp(j)\n",
    "\n",
    "    # insert an item with key k in the heap A ~ $O(logn)$\n",
    "    #   - append the item with key k to the end of the heap: A[n+1] = k\n",
    "    #   - heapifyUp(A, n+1)\n",
    "    def insert(self, key):\n",
    "        self.A.append(key)\n",
    "        self.length += 1\n",
    "        self.size +=1\n",
    "\n",
    "        self.heapifyUp(self.length-1)\n",
    "\n",
    "    def getMax(self):\n",
    "        return P.A[0]\n",
    "    \n",
    "    def pop(self):\n",
    "        self.size -= 1\n",
    "        self.length -= 1\n",
    "        self.A.pop()\n",
    "\n",
    "    # deleteMax:\n",
    "    #  can only easily delete the last element in a dynamic array, but the max of a max_heap is at the root\n",
    "    #    - normally it requres $O(n)$ time by removeing the first element in a dynamic array. can we do it in $logn$ time?\n",
    "    #    - algorithm\n",
    "    #        - swap the max at root node $i = 0$ with the last item at node $n-1$ in heap array, and then delete the last\n",
    "    #        - update heap size by -1\n",
    "    #        - `heapifyDown(A, 0)` after swaping to maintain `max-heap property`\n",
    "    #        - return the deleted node\n",
    "    def deleteMax(self):\n",
    "        self.A[0], self.A[-1] = self.A[-1], self.A[0]\n",
    "        max = self.pop()\n",
    "        self.heapifyDown(0)\n",
    "\n",
    "        return max\n"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Heap Sort\n",
    "\n",
    "HEAPSORT: `heapsort(A)`:\n",
    "- a in-place sorting algorithm that runs in $O(nlogn)$. note that `merge_sort` is also $O(nlogn)$ but requires additional $O(n)$ space.\n",
    "- algorithm\n",
    "  - for i in range(n, -1, 0)\n",
    "  - `deleteMax(A)`\n",
    "   "
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "     \n",
    "## Priority Queue\n",
    "\n",
    "- one of the most popular applications of a heap: as an efficient `priority queue`\n",
    "    - a priority queue is a data structure for maintaining a set `S` of elements, each with an associated value called a key.\n",
    "    - when we use a heap to implement a priority queue, therefore, we often need to store a handle (or key) to the corresponding application object in each heap element. \n",
    "        The exact makeup of the handle (such as a pointer or an integer) depends on the application. \n",
    "        Similarly, we need to store a handle to the corresponding heap element in each application object.\n",
    "    - operations\n",
    "        - `get_max(A)` -> get max of a max-heap A in $O(1)$\n",
    "            - return $A[0]$\n",
    "        - `delete_max(A)` -> same as a max-heap\n",
    "        - `insert(A, k)` -> insert an item with key $k$ in the heap A ~ $O(logn)$\n",
    "            - append the item with key $k$ to the end of the heap: $A[n+1] = k$\n",
    "            - `max_heapify_up(A, n+1)`\n",
    "        - `delete_root()` -> To use a binary heap tree as a priority queue, we will regularly need to delete the root, i.e. remove the node with the highest priority.\n",
    "            - this equals to `delete_max()`\n",
    "\n",
    "        - `delete(A, k)` -> delete an item with key $k$ in the heap A ~ $O(logn)$\n",
    "            - same as the `delete` method in a max_heap\n",
    "        "
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "jupyterbook-py38",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "name": "python",
   "version": "3.8.12 (default, Oct 12 2021, 06:23:56) \n[Clang 10.0.0 ]"
  },
  "orig_nbformat": 4,
  "vscode": {
   "interpreter": {
    "hash": "577791f2da4ac3a13c44f20f0f30021788d0264a6d82c3224a80270143730ffe"
   }
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
